'''
    db-patch applies static and dynamic patches to databases using ODBC or native
    connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
'''


import os

import pyodbc

import dbpatch.options
from dbpatch.loader import Loader
from dbpatch.log import LogFactory
from dbpatch.context import Context
from dbpatch.statement import StatementBuilder


class InvalidConnection(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class Connection(object):

    def __init__(self):
        self._log     = LogFactory().get_logger(Connection.__name__)
        self._options = Context().options
        self._statement_builder = StatementBuilder()
        self._current_db = None
        self._cursor     = None
        self._connection = None

    def __del__(self):
        self.disconnect()

    def create_statement(self, marked_query, values=[], autocommit=True):
        '''
        Create one statement for carry out SQL sentences. This method
        always must be used for execute SQL in any database.
        '''
        return self._statement_builder.create(self, marked_query,
                                              values, autocommit)

    def is_single_database_engine(self):
        '''
        Determine if the database engine support a single database.
        By example SQLite only support a single database, on other
        hand MySQL support several databases.

        Return true if the engine only supports a single database,
        false in other case (by default return false).
        '''
        return False

    def commit(self):
       raise NotImplementedError

    def rollback(self):
        raise NotImplementedError

    def set_current_db(self, current_db):
        self._current_db = current_db

    def connect(self):
        raise NotImplementedError

    def disconnect(self):
        raise NotImplementedError

    def execute(self, query, values=()):
        raise NotImplementedError

    def get_internal_database(self):
        raise NotImplementedError

    def get_all_databases(self):
        raise NotImplementedError

    def get_row(self):
        raise NotImplementedError

    def get_rows(self):
        raise NotImplementedError

    def get_rows_count(self):
        raise NotImplementedError


class DbApi2Connection(Connection):
    
    def commit(self):
        if not self._connection:
            return
        self._connection.commit()

    def rollback(self):
        if not self._connection:
            return
        self._connection.rollback()

    def disconnect(self):
        if self._cursor:
            self._cursor.close()

    def execute(self, query, values=()):
        if not self._connection:
            raise InvalidConnection('Not connected to database')

        ok = True
        self._cursor = self._connection.cursor()
        try:
            try:
                self._cursor.execute(query, values)
                self._log.debug('Afected [%d] rows by [%s...]' \
                                    % (self.get_rows_count(),
                                       query[0:50].replace('\n', ' ')))
            except Exception, e:
                self._log.error('Error executing [%s]. '
                                'Error: %s' % (query, str(e)))
                ok = False
        finally:
            self._connection.commit()

        return ok

    def get_row(self):
        if not self._cursor:
            return None

        return self._cursor.fetchone()

    def get_rows(self):
        if not self._cursor:
            return None

        return self._cursor.fetchall()

    def get_rows_count(self):
        if not self._cursor:
            return -1

        return self._cursor.rowcount



class ConnectionFactory(Loader):

    CONN_CLASSES = {}

    def __init__(self, extra_path=None):
        import dbpatch.connection
        base_path = os.path.dirname(dbpatch.connection.__file__)
        Loader.__init__(self, base_path, 'ConnectionFactory', extra_path)

    def _create_native(self):
        Context().set_dbapi_conn_type()
        the_id = Context().options.get_option(
            dbpatch.options.DB_ENGINE).lower()
        self._log.debug('Creating native connection [%s]' % the_id)
        try:
            return ConnectionFactory.CONN_CLASSES[the_id]()
        except KeyError, invalid_key:
            raise InvalidConnection(
                'Unable create a connection of type [%s]' % invalid_key)
        
    def create(self):
        if not Context().options.get_option(dbpatch.options.CONN_NATIVE):
            Context().set_odbc_conn_type()
            import dbpatch.connection.odbc as odbc
            return odbc.OdbcConnection()

        return self._create_native()

    def get_dict_classes(self):
        return ConnectionFactory.CONN_CLASSES

    @classmethod
    def register_connection(cls, the_class):
        try:
            ConnectionFactory.CONN_CLASSES[the_class.NAME] = the_class
        except Exception, error:
            Loader.LOG.error('Registering %s [%s]' % (the_class, error))
