'''
    db-patch applies static and dynamic patches to databases using ODBC or native
    connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
'''


import os
import configobj

import pyodbc

import dbpatch.options
from dbpatch.connection import Connection, ConnectionFactory
from dbpatch.loader import Loader
from dbpatch.log import LogFactory
from dbpatch.statement import StatementBuilder
from dbpatch.context import Context


class InvalidOdbcQueryBuilder(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)

class UndefinedODBCdriver(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)



class OdbcQueryBuilder(object):

    def __init__(self):
        self._log     = LogFactory().get_logger(OdbcQueryBuilder.__name__)
        self._options = Context().options
        self._odbc_instini = self._options.get_option(
            dbpatch.options.ODBC_INST_INI) 
        self._drivers = configobj.ConfigObj(self._odbc_instini).sections

    def _replace_keywords(self, connection, keywords={}):
        if len(keywords.keys()) == 0:
            return connection

        for standard, customized in keywords.items():
            connection = connection.replace('%s=' % standard,
                                            '%s=' % customized)

        return connection


    def __check_options(self, options, mode):
        for option_name in options:
            if self._options.get_option(option_name):
                continue
            raise dbpatch.options.RequiredOption(
                'Required option [%s] for %s connection' % (option_name,
                                                            mode))

    def __option_to_conn(self, keyword, opt_name):
        value = self._options.get_option(opt_name)
        if not value:
            return ''

        return '%s=%s;' % (keyword, value)

    def __driver_to_conn(self, keyword, opt_name):
        driver = self._options.get_option(opt_name)

        if driver in self._drivers:
            return '%s={%s};' % (keyword, driver)
            
        raise UndefinedODBCdriver(
            'Undefined driver [%s] in odbcinst.ini file [%s]' % \
                (driver, self._odbc_instini))


    def is_single_database_engine(self):
        """
        Determine if the database engine support a single database.
        By example SQLite only support a single database, on other
        hand MySQL support several databases.

        Return true if the engine only supports a single database,
        false in other case (by default return false).
        """
        return False

    def get_mandatory_dsn_options(self):
        """
        Return a dictionary (option identifier (defined in
        dbpatch.options), option name) with the mandatory options used
        when a DSN connection is used.

        The child classes can overwrite this method which is used when
        the connection string is generated.
        """
        return {dbpatch.options.DB_PASS  : 'PASSWORD',
                dbpatch.options.DB_USER  : 'USER',
                dbpatch.options.ODBC_DSN : 'DSN',
                }

    def get_mandatory_non_dsn_options(self):
        """
        Return a dictionary (option identifier (defined in
        dbpatch.options), option name) with the mandatory options used
        when a non DSN connection is used.

        The child classes can overwrite this method which is used when
        the connection string is generated.
        """
        return {dbpatch.options.DB_PASS     : 'PASSWORD',
                dbpatch.options.DB_USER     : 'USER',
                dbpatch.options.DB_HOST     : 'SERVER',
                dbpatch.options.ODBC_DRIVER : 'DRIVER',
                }

    def get_optional_options(self):
        """
        Return a dictionary {option identifier (defined in
        dbpatch.options) : option name} with the optional options
        used when a string connection is generated.

        The child classes can overwrite this method
        """
        return {dbpatch.options.DB_PORT     : 'PORT',
                dbpatch.options.DB_CHARSET  : 'CHARSET',
                }

    def get_internal_database(self):
        """
        If it is not required to use the internal database for recover
        the list of databases the method should return None. In other
        case, it should return the internal database name.
        """
        return None

    def get_all_databases_query(self):
        """
        Return the query that the database engine uses for recover all
        the databases.
        """
        raise NotImplementedError

    def get_customized_keywords(self):
        """
        Return a dictionary {default name : customized name} that
        allow change the connection string identifiers.

        By example, if it is necessary change the USER=... by UID=... this
        method should return the next dictionary: {'USER':'UID'}.
        """
        return {}

    def get_connection_string(self, db_name=None):
        """
        Return the connection string required to connect via ODBC.
        
        It takes into account the next methods for create the
        connection string.

          get_mandatory_dsn_options
          get_mandatory_non_dsn_options
          get_optional_options
          get_customized_keywords

        Child classes can overwrite this method.
        """
        if self._options.get_option(dbpatch.options.ODBC_FULL_CONN):
            return self._options.get_option(dbpatch.options.ODBC_FULL_CONN)

        driver  = self._options.get_option(dbpatch.options.ODBC_DRIVER)
        extra   = self._options.get_option(dbpatch.options.ODBC_EXTRA)
        dsn     = self._options.get_option(dbpatch.options.ODBC_DSN)
        server  = self._options.get_option(dbpatch.options.DB_HOST)
        port    = self._options.get_option(dbpatch.options.DB_PORT)
        user    = self._options.get_option(dbpatch.options.DB_USER)
        pwd     = self._options.get_option(dbpatch.options.DB_PASS)
        charset = self._options.get_option(dbpatch.options.DB_CHARSET)
        connection = ''
        
        if dsn:
            options = self.get_mandatory_dsn_options()
            mode    = 'DSN'

        else:
            options = self.get_mandatory_non_dsn_options()
            mode    = 'non-DSN'

        self.__check_options(options.keys(), mode)

        for option_name, keyword in options.items():
            if option_name == dbpatch.options.ODBC_DRIVER:
                connection += self.__driver_to_conn(keyword, option_name)
            else:
                connection += self.__option_to_conn(keyword, option_name)

        for option_name, keyword in self.get_optional_options().items():
            connection += self.__option_to_conn(keyword, option_name)

        if extra:
            connection += extra

        if db_name:
            connection += 'DATABASE=%s;' % db_name

        connection = self._replace_keywords(connection,
                                            self.get_customized_keywords())

        self._log.debug('Connection chain [%s]' % connection)
        return connection



class OdbcQueryBuilderFactory(Loader):

    QBUILDER_CLASSES = {}

    def __init__(self, extra_path=None):
        import dbpatch.connection.odbc
        base_path = os.path.dirname(dbpatch.connection.odbc.__file__)
        Loader.__init__(self, base_path,
                        'OdbcQueryBuilderFactory',
                        extra_path)

    def create(self):
        the_id = Context().options.get_option(
            dbpatch.options.DB_ENGINE).lower()
        self._log.debug('Creating odbc builder for [%s]' % the_id)
        try:
            return OdbcQueryBuilderFactory.QBUILDER_CLASSES[the_id]()
        except KeyError, invalid_key:
            raise InvalidOdbcQueryBuilder(
                'Unable create a query builder for database engine [%s]' % invalid_key)

    def get_dict_classes(self):
        return OdbcQueryBuilderFactory.QBUILDER_CLASSES

    @classmethod
    def register(cls, the_class):
        try:
            OdbcQueryBuilderFactory.QBUILDER_CLASSES[the_class.NAME] = the_class
        except Exception, error:
            Loader.LOG.error('Registering %s [%s]' % (the_class, error))


class OdbcConnection(Connection):

    def __init__(self):
        Connection.__init__(self)
        self._log = LogFactory().get_logger(Connection.__name__)
        extra_path = self._options.get_option(dbpatch.options.ODBC_PLUGINS)
        self._odbc_builder = OdbcQueryBuilderFactory(extra_path).create()
        self._statement_builder = StatementBuilder()

    def commit(self):
        if not self._connection:
            return
        self._connection.commit()

    def rollback(self):
        if not self._connection:
            return
        self._connection.rollback()

    def _is_connected_to(self, db_name):
        if self._connection and self._current_db == db_name :
            return True

        return False

    def _connect(self):
        current_db = Context().current_db
        self._connection = pyodbc.connect(
            self._odbc_builder.get_connection_string(current_db))
        self._current_db = current_db

    def connect(self):
        if self._is_connected_to(Context().current_db):
            return

        self._connect()

    def disconnect(self):
        if self._cursor:
            self._cursor.close()

        if self._connection:
            self._connection.close()

    def execute(self, query, values=()):
        self._cursor = self._connection.execute(query, values)
        return True

    def get_internal_database(self):
        if self._odbc_builder.is_single_database_engine():
            return self._current_db

        return self._odbc_builder.get_internal_database()

    def get_all_databases(self):
        # The engine does not support more than one database. The
        # database list passed by parameter must be used for this
        # purpose
        if self._odbc_builder.is_single_database_engine():
            return self._options.get_option(
                dbpatch.options.DATABASES).split(',')

        # Some databases require to be connected to the internal
        # database for recover the list of working databases
        internal = self.get_internal_database()
        if not self._is_connected_to(internal):
            Context().current_db = internal
            self.connect()

        # If the database does not have query for recover the
        # databases list only the current db will be used
        if not self._odbc_builder.get_all_databases_query():
            return [self._current_db]

        self.execute(self._odbc_builder.get_all_databases_query())
        rows = self.get_rows()
        dbs  = []
        for row in rows:
            dbs.append(row[0])

        self._log.trace('All databases:  %s' % str(dbs))
        return dbs

    def get_row(self):
        if not self._cursor:
            return None

        return self._cursor.fetchone()

    def get_rows(self):
        if not self._cursor:
            return []

        return self._cursor.fetchall()

    def get_rows_count(self):
        if not self._cursor:
            return 0

        return self._cursor.rowcount

