'''
    db-patch applies static and dynamic patches to databases using ODBC or native
    connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
'''


# -*- tab-width:4; coding:utf-8 -*-

import os
import os.path
import sys
import imp
import inspect
from dbpatch.log import LogFactory


def load_module(full_name):
    (path, name) = os.path.split(full_name)
    (name, ext) = os.path.splitext(name)

    (fd, filename, data) = imp.find_module(name, [path])
    return imp.load_module(name, fd, filename, data)

def get_classes(module):
    return inspect.getmembers(module, inspect.isclass)

def new_class(module, class_name, **kargs):
    the_class = getattr(module, class_name)
    return the_class(**kargs)

def new_class_child_of(module, parent_class, **kargs):
    loaded_module = load_module(module)
    the_classes   = get_classes(loaded_module)
    hierarchies   = []
    for name, loaded_class in the_classes:
        hierarchies.append((name, inspect.getmro(loaded_class)))

    for name, hierarchy in hierarchies:
        if len(hierarchy) < 2:
            continue

        for child in hierarchy:
            if str(child) != parent_class:
                continue

            return getattr(loaded_module, name)(**kargs)
    
    return None


class Loader(object):

    DICT_CLASSES = {}
    LOG = LogFactory().get_logger('Loader')

    def __init__(self, base_path, log_name, extra_path=None):
        self._paths = [base_path]
        self._log   = LogFactory().get_logger(log_name)

        if extra_path:
            full_path = os.path.realpath(os.path.expanduser(extra_path))
            if os.path.isdir(full_path):
                self._paths.append(extra_path)
                self._log.trace('Using extra path [%s]' % full_path)

        self.__update_sys_path()
        self.__load_class()
        self.__restore_sys_path()

        self._log.trace('Loaded: %s' % str(self.get_dict_classes().keys()))

    def create(self, class_identifier, **args):
        self._log.trace('Creating [%s]' % class_identifier)
        return Loader.DICT_CLASSES[class_identifier](**args)

    def get_dict_classes(self):
        return Loader.DICT_CLASSES

    @classmethod
    def register(cls, the_class):
        try:
            Loader.DICT_CLASSES[the_class.NAME] = the_class
        except Exception, error:
            Loader.LOG.error('Registering %s [%s]' % (the_class, error))
            
    def __update_sys_path(self):
        for new_path in self._paths:
            if new_path in sys.path:
                continue
            self._log.trace('Adding [%s] to sys.path' % new_path)
            sys.path.append(new_path)

    def __restore_sys_path(self):
        for new_path in self._paths:
            if new_path not in sys.path:
                continue
            self._log.trace('Removing [%s] of sys.path' % new_path)
            sys.path.remove(new_path)

    def __recover_class_files(self):
        class_files = []

        for class_path in self._paths:
            if not os.path.isdir(class_path):
                continue

            for name in os.listdir(class_path):
                if name.startswith('.') or \
                       name.startswith('#') or \
                       name.startswith('__') or \
                       not name.endswith('.py') :
                    continue

                class_files.append(name)

        return class_files


    def __load_class(self):
        class_files = self.__recover_class_files()

        for the_file in class_files:
            module = __import__(the_file[:-3])
            try:
                module.register()
            except Exception, error:
                self._log.error('Required register() method!! in '
                                '%s [%s]' % (the_file, error))


