'''
    db-patch applies static and dynamic patches to databases using ODBC or native
    connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
'''


import sys
import os
import subprocess
from ConfigParser import ConfigParser
from optparse import OptionParser, OptionGroup

import dbpatch
from dbpatch.context import Context


# Options for ODBC connections
ODBC_DRIVER     = 'odbc_driver'
ODBC_DSN        = 'odbc_dsn'
ODBC_EXTRA      = 'odbc_extra'
ODBC_FULL_CONN  = 'odbc_full_conn'
ODBC_INST_INI   = 'odbc_instini'
ODBC_PLUGINS    = 'odbc_plugins'
CONN_PLUGINS    = 'connection_plugins'
CONN_NATIVE     = 'native'

# Options for database
DB_USER    = 'db_user'
DB_PASS    = 'db_pass'
DB_HOST    = 'db_host'
DB_PORT    = 'db_port'
DB_ENGINE  = 'db_engine'
DB_SOCKET  = 'db_socket'
DB_CHARSET = 'db_charset'
DB_CHARSET_POLICY = 'db_charset_policy'

# Options for filter the database list
EXCLUDE   = 'exclude'
INCLUDE   = 'include'
DATABASES = 'databases'

# Patch options
PATCH_FILE    = 'patch_file'
PATCH_TYPE    = 'patch_type'
PATCH_PLUGINS = 'patch_plugins'

# Generated SQL options
SQL_CLEAN    = 'clean'
SQL_GEN_DIR  = 'generated_dir'
SQL_GEN_FILE = 'gen_file'
SQL_SHOW     = 'show_sql'

# Run options
DRY_RUN  = 'dry_run'
VERBOSE  = 'verbose'
PRE_CMD  = 'pre_cmd'
POST_CMD = 'post_cmd'

# Misc options
EXTRA_OPTS = 'extra_opts'
ARGS_FILE  = 'args_file'


class ConfigurationFileException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class RequiredOption(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)



class Option(object):

    def __init__(self):
        pass

    def add_option(self, parser):
        pass

    def get_destiny(self):
        return self._destiny
    

class LongOption(Option):

    def __init__(self, long_opts, description, default_value, destiny):
        Option.__init__(self)
        self._long_opts  = long_opts
        self._description = description
        self._default = default_value
        self._destiny = destiny

    def add_option(self, parser):
        parser.add_option(self._long_opts,
                          help    = self._description,
                          dest    = self._destiny,
                          default = self._default)


class ShortLongOption(Option):

    def __init__(self, short_opts, long_opts, description,
                 default_value, destiny):
        Option.__init__(self)
        self._short_opts = short_opts
        self._long_opts  = long_opts
        self._description = description
        self._default = default_value
        self._destiny = destiny

    def add_option(self, parser):
        parser.add_option(self._short_opts,
                          self._long_opts,
                          help    = self._description,
                          dest    = self._destiny,
                          default = self._default)

class BooleanOption(Option):

    def __init__(self, short_opts, long_opts, description,
                 default_value, destiny, behaviour):
        Option.__init__(self)
        self._short_opts = short_opts
        self._long_opts  = long_opts
        self._description = description
        self._default = default_value
        self._destiny = destiny
        self._behaviour = behaviour

    def add_option(self, parser):
        parser.add_option(self._short_opts,
                          self._long_opts,
                          help    = self._description,
                          dest    = self._destiny,
                          default = self._default,
                          action  = self._behaviour)


class Group(object):

    def __init__(self, parser):
        self._group  = OptionGroup(parser, self.get_title())
        self.__populate_group()
        
    def __populate_group(self):
        for option in self.get_raw_options():
            option.add_option(self._group)

    def get_title(self):
        return 'The title'

    def get_raw_options(self):
        return []

    def get_group(self):
        return self._group
    


class OdbcOptions(Group):

    ODBC_TITLE = 'Connection and ODBC options'
    
    ODBC_OPTS = [
        LongOption('--odbc-driver',
                   'Define the ODBC driver to use. '
                   'It must be defined in the odbcinst.ini',
                   None,
                   ODBC_DRIVER),
        LongOption('--odbc-dsn',
                   'Define a DSN for ODBC connections',
                   None,
                   ODBC_DSN),
        LongOption('--odbc-extra',
                   'Provide extra options for make the ODBC '
                   'connection string',
                   None,
                   ODBC_EXTRA),
        LongOption('--odbc-full-connection',
                   'Define the full ODBC connection string',
                   None,
                   ODBC_FULL_CONN),
        LongOption('--odbc-inst-ini',
                   'Full path to odbcinst.ini. '
                   'By default: /etc/odbcinst.ini',
                   '/etc/odbcinst.ini',
                   ODBC_INST_INI),
        BooleanOption('-N', '--native',
                      'Use a native driver for database connection. '
                      'By default uses ODBC (requires to configure '
                      'the odbcinst.ini file)',
                      False,
                      CONN_NATIVE,
                      'store_true'),
        LongOption('--odbc-plugins',
                   'Define the ODBC query builder path '
                   'for locate custom ODBC connections.'
                   'By default: ~/.dbpatch/odbc',
                   '~/.dbpatch/odbc',
                   ODBC_PLUGINS),
        LongOption('--connection-plugins',
                   'Extra path for locate custom native connections. '
                   'By default: ~/.dbpatch/connection',
                   '~/.dbpatch/connection',
                   CONN_PLUGINS),

        ]

    def __init__(self, parser):
        Group.__init__(self, parser)

    def get_title(self):
        return OdbcOptions.ODBC_TITLE

    def get_raw_options(self):
        return OdbcOptions.ODBC_OPTS[:]



class DatabaseOptions(Group):

    DB_TITLE = 'Database options'
    
    DB_OPTS = [
        ShortLongOption('-u', '--db-user',
                        'Database user. Mandatory',
                        None,
                        DB_USER),
        ShortLongOption('-p', '--db-pass',
                        'Database password. Mandatory',
                        None,
                        DB_PASS),
        ShortLongOption('-H', '--db-host',
                        'Database host. By default "localhost"',
                        'localhost',
                        DB_HOST),
        ShortLongOption('-P', '--db-port',
                        'Database port',
                        None,
                        DB_PORT),
        ShortLongOption('-E', '--db-engine',
                        'Database engine (MySQL, PostgreSQL, etc). Mandatory',
                        None,
                        DB_ENGINE),
        ShortLongOption('-C', '--db-charset',
                        'The charset used for database connections.By default: ascii',
                        'ascii',
                        DB_CHARSET),
        ShortLongOption('-s', '--db-socket',
                        'Database Unix socket',
                        None,
                        DB_SOCKET),
        ShortLongOption('-e', '--exclude',
                        'Regular expression for exclude databases.'
                        'Mandatory to use python regular expressions (re module). '
                        'Example: -e "a[a-z]+"',
                        None,
                        EXCLUDE),
        ShortLongOption('-i', '--include',
                        'Regular expression for include databases.'
                        'See python re module',
                        None,
                        INCLUDE),
        ShortLongOption('-d', '--databases',
                        'List of databases to patch.'
                        'Examples: -d W24_WE1 or -d W24_WE1,W24_WH6',
                        None,
                        DATABASES),
        LongOption('--db-charset-policy',
                   'The policy to apply when the charset can not full replace strings. '
                   'Posible values: [ignore|replace|xmlcharrefreplace]. '
                   'By default: ignore',
                   'ignore',
                   DB_CHARSET_POLICY)
        ]

    def __init__(self, parser):
        Group.__init__(self, parser)

    def get_title(self):
        return DatabaseOptions.DB_TITLE

    def get_raw_options(self):
        return DatabaseOptions.DB_OPTS[:]



class PatchOptions(Group):

    PATCH_TITLE = 'Patch options'
    
    PATCH_OPTS = [
        ShortLongOption(
            '-f', '--file',
            'File that contains the patch. The patch executor uses the extension for '
            'runs each patch. For the ".sql" patches the mydb keyword will be '
            'replaced by each database name. Mandatory',
            None,
            PATCH_FILE),
        LongOption('--patch-plugins',
                   'Extra path for locate customized patchs. '
                   'By default: ~/.dbpatch/patch',
                   '~/.dbpatch/patch',
                   PATCH_PLUGINS),
        LongOption('--patch-type',
                   'Patch type. Supported: sql, python. '
                   'Extensible via patch plugins (see --patch-plugins).'
                   'By default the patch type will be determined by the file extension',
                   None,
                   PATCH_TYPE),
        ]

    def __init__(self, parser):
        Group.__init__(self, parser)

    def get_title(self):
        return PatchOptions.PATCH_TITLE

    def get_raw_options(self):
        return PatchOptions.PATCH_OPTS[:]



class SqlOptions(Group):

    SQL_TITLE = 'SQL tracing options'
    
    SQL_OPTS = [
        BooleanOption('-c', '--clean',
                      'Avoid that all the SQL executed will be stored',
                      False,
                      SQL_CLEAN,
                      'store_true'),
        LongOption('--gen-dir',
                   'Directory where the SQL sentences for each patch will be stored. '
                   'By default: sql_sentences',
                   'sql_sentences',
                   SQL_GEN_DIR),
        LongOption('--gen-file',
                   'Determines the name of the file where the generated SQL is stored.'
                   'The "HOST" flag will be replaced by the hostname.'
                   'The "NAME" flag will be replaced by the patch name without extension.'
                   'The "TIME" flag will be replaced by a timestamp. '
                   'By default: "HOST_NAME_TIME.sql"',
                   'HOST_NAME_TIME.sql',
                   SQL_GEN_FILE),
        BooleanOption('-S', '--show-sql',
                      'Show the generated SQL sentences in the output',
                      False,
                      SQL_SHOW,
                      'store_true'),
        ]

    def __init__(self, parser):
        Group.__init__(self, parser)

    def get_title(self):
        return SqlOptions.SQL_TITLE

    def get_raw_options(self):
        return SqlOptions.SQL_OPTS[:]



class RunOptions(Group):

    RUN_TITLE = 'Run options'
    
    RUN_OPTS = [
        BooleanOption('-t', '--dry-run',
                      'Generate all the SQL sentences without modify the database',
                      False,
                      DRY_RUN,
                      'store_true'),
        BooleanOption('-v', '--verbose',
                      'Increase the verbose level', 0, VERBOSE, 'count'),
        LongOption('--pre-cmd',
                   "Command to be executed before of execute the patch. "
                   "By example --pre-cmd 'find -name \"*.pyc\" -delete' ",
                   None,
                   PRE_CMD),
        LongOption('--post-cmd',
                   "Command to be executed after of execute the patch. "
                   "By example --post-cmd "
                   "'tar -czf applied.tgz sql_generated/; "
                   "rm -rf sql_generated/' ",
                   None,
                   POST_CMD)
        ]

    def __init__(self, parser):
        Group.__init__(self, parser)

    def get_title(self):
        return RunOptions.RUN_TITLE

    def get_raw_options(self):
        return RunOptions.RUN_OPTS[:]



class MiscOptions(Group):

    MISC_TITLE = 'Extra options'
    
    MISC_OPTS = [
        LongOption('--extra-opts',
                   'Allow use a configuration file for introduce new '
                   'options. The options will be stored as '
                   '"section_name.option_name"',
                   None,
                   EXTRA_OPTS),
        LongOption('--args-file',
                   'Allow use a file for introduce the arguments.'
                   'The command line options will overwrite the options '
                   'defined in the file',
                   None,
                   ARGS_FILE),
        ]

    def __init__(self, parser):
        Group.__init__(self, parser)

    def get_title(self):
        return MiscOptions.MISC_TITLE

    def get_raw_options(self):
        return MiscOptions.MISC_OPTS[:]



class Options:

    DESCRIPTION = '%s v%s' % (dbpatch.SHORT_DESCRIPTION,
                              dbpatch.VERSION)

    def __init__(self):
        self._parser  = OptionParser(description = Options.DESCRIPTION)
        self._options = {}

        db    = DatabaseOptions(self._parser)
        patch = PatchOptions(self._parser)
        odbc  = OdbcOptions(self._parser)
        sql   = SqlOptions(self._parser)
        run   = RunOptions(self._parser)
        misc  = MiscOptions(self._parser)
        self._groups = [db, patch, odbc, sql, run, misc]

        for opt_group in self._groups :
            self._parser.add_option_group(opt_group.get_group())

        self._databases = {'include'   : [],
                           'exclude'   : [],
                           'databases' : []}

    def get_terminal_size(self):
        output = subprocess.Popen('stty size', shell=True,
                                  stdout=subprocess.PIPE,
                                  close_fds=True)
        out = output.communicate()[0]
        if output.returncode > 0:
            return (80, 80)
        
        size = out.split()
        return (int(size[0]), int(size[1]))

    def parse(self):
        options, args = self._parser.parse_args()

        args_file = getattr(options, ARGS_FILE)
        if args_file:
            args = open(args_file).read().split()
            arguments = sys.argv[1:]
            arguments.extend(args)
            options, args = self._parser.parse_args(arguments)

        for group in self._groups:
            for opt in group.get_raw_options():
                key = opt.get_destiny()
                self._options[key] = getattr(options, key)

        for name in [INCLUDE, EXCLUDE, DATABASES]:
            if not getattr(options, name):
                continue
            self._databases[name] = getattr(options,name).split(',')

        if self._options[EXTRA_OPTS]:
            self.__populate_from_file(self._options[EXTRA_OPTS])

        Context().options = self

    def get_options(self):
        return self._options.copy()

    def get_option(self, the_option):
        return self._options[the_option]

    def set_options(self, options):
        self._options.update(options)

    def get_mandatory_option(self, the_option):
        if not the_option in self._options.keys():
            raise RequiredOption('The option [%s] is mandatory' % the_option)

        value = self._options[the_option]

        if not value:
            raise RequiredOption('The option [%s] is mandatory. '
                                 'The value cannot be None' % the_option)

        return value

    def get_include_dbs(self):
        return self._databases[INCLUDE]

    def get_exclude_dbs(self):
        return self._databases[EXCLUDE]

    def get_only_dbs(self):
        return self._databases[DATABASES]

    def __populate_from_file(self, the_file):
        if not os.path.isfile(the_file):
            raise ConfigurationFileException('The file [%s] does not exist' \
                                                 % the_file)
        config = ConfigParser()
        config.read(the_file)

        for section in config.sections():
            for opt in config.options(section):
                key = '%s.%s' % (section, opt)
                self._options[key] = config.get(section, opt)

