'''
    db-patch applies static and dynamic patches to databases using ODBC or native
    connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
'''


import os

import dbpatch.options
from dbpatch.command import Command
from dbpatch.loader import Loader
from dbpatch.connection import ConnectionFactory
from dbpatch.dbfilter import DbFilter
from dbpatch.log import LogFactory
from dbpatch.context import Context


class InvalidPatchFile(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class PatchFactory(Loader):

    PATCH_CLASSES = {}

    def __init__(self, extra_path=None):
        import dbpatch.patch
        base_path = os.path.dirname(dbpatch.patch.__file__)
        Loader.__init__(self, base_path, 'PatchFactory', extra_path)

    def create(self, class_identifier):
        Loader.LOG.debug('Creating patch of type [%s]' % class_identifier)
        try:
            return PatchFactory.PATCH_CLASSES[class_identifier]()
        except KeyError, invalid_key:
            raise InvalidPatchFile(
                'Unable create a patch for patch extension %s. ' \
                'Registered %s .' \
                    'Class identifier %s' % \
                    (invalid_key,
                     str(PatchFactory.PATCH_CLASSES.keys()),
                     class_identifier))

    def get_dict_classes(self):
        return PatchFactory.PATCH_CLASSES

    @classmethod
    def register(cls, the_class):
        if the_class.EXTENSION in PatchFactory.PATCH_CLASSES.keys():
            raise InvalidPatchFile(
                'The extension [%s] is already used by other plugin' \
                    % the_class.EXTENSION)

        try:
            PatchFactory.PATCH_CLASSES[the_class.EXTENSION] = the_class
        except Exception, error:
            Loader.LOG.error('Registering %s [%s]' % (the_class, error))



class Patch:

    def __init__(self):
        self._ctx  = Context()
        self._options = self._ctx.options
        extra_path = self._options.get_option(dbpatch.options.CONN_PLUGINS)
        self._connection = ConnectionFactory(extra_path).create()
        self._error = ''

    def create_statement(self, marked_query, values=(), autocommit=True):
        '''
        Create one statement for carry out SQL sentences. This method
        always must be used for execute SQL in any database.
        '''
        return self._connection.create_statement(marked_query, values)

    def encode(self, the_string):
        '''
        Encode the string passed by parameter taking into account the
        charset and the charset policy defined for the patch (through
        command line options).
        '''
        return the_string.encode(
            self._options.get_option(dbpatch.options.DB_CHARSET),
            self._options.get_option(dbpatch.options.DB_CHARSET_POLICY))

    def connect_to(self, db_name):
        '''
        Connect with the database passed by parameter. The patch
        context is updated.
        '''
        self._ctx.current_db = db_name
        return self._connection.connect()

    def get_databases(self):
        '''
        Return the list of databases taking into account the
        include/exclude options.
        '''
        all_dbs = self._connection.get_all_databases()
        exclude = self._options.get_option(dbpatch.options.EXCLUDE)
        include = self._options.get_option(dbpatch.options.INCLUDE)
        only    = self._options.get_only_dbs()
        
        return DbFilter(all_dbs, exclude, include, only).apply_filter()

    def get_error(self):
        return self._error

    def set_error(self, error):
        '''
        The patchs can set an error message if in any stage there
        are errors. This error it will be shown to final user when
        the patch process stops.
        '''
        self._error = error

    def get_options(self):
        return self._options

    def pre_execute(self):
        '''
        This method it will be executed once before execute the
        patch in all databases.
        Must return true if the excution is ok, false in other case
        '''
        return True

    def before_database(self, db_name):
        '''
        This method it will be executed before of execute the patch in
        each database.
        '''
        pass

    def execute(self, db_name):
        '''
        This method it will be executed in each database. The database
        list is determined through the options -e, -i and -d.
        Must return true if the excution is ok, false in other case
        '''
        return True

    def after_database(self, db_name):
        '''
        This method it will be executed after of execute the patch in
        each database.
        '''
        pass

    def post_execute(self):
        '''
        This method it will be executed once after execute the
        patch in all databases.
        Must return true if the excution is ok, false in other case
        '''
        return True



class PatchExecutor(object):

    MANDATORY_OPTIONS = ['db_user', 'db_pass', 'db_engine', 'patch_file']

    def __init__(self):
        self._ctx = Context()
        self._log = LogFactory().get_logger(PatchExecutor.__name__)
        self._options = self._ctx.options
        self._error   = ''
        self.__check_mandatory_options()

    def __check_mandatory_options(self):
        for mandatory in PatchExecutor.MANDATORY_OPTIONS:
            if self._options.get_option(mandatory):
                continue
            raise dbpatch.options.RequiredOption(
                'Required %s option' % mandatory.upper())


    def create_patch(self):
        engine = self._options.get_option(dbpatch.options.DB_ENGINE).lower()
        if self._options.get_option(dbpatch.options.PATCH_TYPE):
            patch_type = self._options.get_option(
                dbpatch.options.PATCH_TYPE).lower()
        else:
            patch_type = self._options.get_option(
                dbpatch.options.PATCH_FILE).split('.')[1]

        extra_path = self._options.get_option(dbpatch.options.PATCH_PLUGINS)
        patch = PatchFactory(extra_path).create(patch_type)
        self._ctx.current_patch = self._options.get_option(
            dbpatch.options.PATCH_FILE)
        return patch

    def _execute_command(self, cmd_opt, cmd_prefix):
        cmd = self._options.get_option(cmd_opt)
        if not cmd:
            return True

        command = Command(cmd)
        ok  = command.run()
        if ok != 0:
            self._log.error(
                'Fail executing the %s command [%s]. Out[%s]. Err[%s]' % \
                    (cmd_prefix, cmd, command.stdout, command.stderr))
            return False
        
        self._log.trace('Executed %s command [%s]' % (cmd_prefix, cmd))
        return True

    def _execute_pre_command(self):
        return self._execute_command(dbpatch.options.PRE_CMD, 'pre')

    def _execute_post_command(self):
        return self._execute_command(dbpatch.options.POST_CMD, 'post')

    def pre_execute_stage(self, patch):
        ok_pre_cmd = self._execute_pre_command()
        if not ok_pre_cmd:
            return ok_pre_cmd

        self._log.debug('Started pre_execute stage')
        ok = patch.pre_execute()
        self._log.debug('Finished pre_execute stage [Success:%s]' % str(ok))
        return ok

    def execute_stage(self, patch):
        self._log.debug('Started execute stage')

        for db in patch.get_databases():
            patch.before_database(db)

            self._ctx.current_db = db

            ok = patch.execute(db)
            self._log.debug('Finished execute stage on database [%s] ' \
                                '[Success:%s]' % (db, str(ok)))
            if not ok:
                return False

            patch.after_database(db)

        return True

    def post_execute_stage(self, patch):
        self._log.debug('Started post_execute stage')
        ok = patch.pre_execute()
        self._log.debug('Finished post_execute stage [Success:%s]' % str(ok))
        return ok

    def run(self):
        patch = self.create_patch()

        if not self._execute_pre_command():
            self._log.error('FAIL! Executing pre command. Aborted')
            return
        
        if not self.pre_execute_stage(patch):
            self._log.error('FAIL! Pre-execute stage error [%s]' % str(patch.get_error()))
            return

        if not self.execute_stage(patch):
            self._log.error('FAIL! Execute stage error [%s]' % str(patch.get_error()))
            return

        if not self.post_execute_stage(patch):
            self._log.error('FAIL! Post-execute stage error [%s]' % str(patch.get_error()))
            return

        if not self._execute_post_command():
            self._log.error('FAIL! Executing post command. '
                            'Finished without execute the post command')
            return

        self._ctx.set_end_time()
        self._log.info('Finished :-)')
