'''
    db-patch applies static and dynamic patches to databases using ODBC or native
    connectors.
    Copyright (C) 2012, 2013  Miguel A. Martinez Pinedo

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
'''


import re

import codecs

import dbpatch.options
from dbpatch.log import LogFactory
import dbpatch.patch

try:
    import sqlparse
except ImportError:
    print("\n\tFAIL!! Required the package: python-sqlparse\n")
    import sys
    sys.exit(1)


class Query:

    DB_TEMPLATE_NAME = 'mydb'

    def __init__(self, sql, n_query):
        self._sql     = sql
        self._n_query = n_query

    def __repr__(self):
        return "\nQuery Number:%i\nSQL : [%s]" % (self._n_query, self._sql)

    def get_number_query(self):
        return self._n_query

    def get_sql(self):
        return self._sql

    def get_sql_for_database(self, db_name):
        return self._sql.replace(Query.DB_TEMPLATE_NAME, db_name)



class SQLSplitter:

    def __init__(self, patch_file_name):
        self._path_file_name = patch_file_name
        self._queries   = []
        self._file_desc = open(self._path_file_name, 'r')
        queries = sqlparse.split(self._file_desc.read())
        self._file_desc.close()

        for i in range(0, len(queries)):
            without_comments = SQLSplitter.clean_comments(queries[i])

            if len(without_comments.strip()) == 0:
                continue

            self._queries.append(Query(without_comments, i))
           
    @staticmethod
    def clean_comments(sql):
        return re.sub('--[-]*[\s*\w*]*\n', '', sql)

    def get_queries(self):
        return self._queries



class SQLPatch(dbpatch.patch.Patch):
    '''
    Split the content of a file in queries, and execute them in the
    database.
    
    Replace the keyword 'mydb' for each database name
    '''

    EXTENSION = 'sql'

    def __init__(self):
        dbpatch.patch.Patch.__init__(self)
        self._log = LogFactory().get_logger(SQLPatch.__name__)
        patch_file = self._options.get_option(dbpatch.options.PATCH_FILE)
        self._queries = SQLSplitter(patch_file).get_queries()

    def execute(self, db_name):
        self.connect_to(db_name)
        for query in self._queries:
            sql = query.get_sql_for_database(db_name)
            sql = self.encode(sql)
            statement = self.create_statement(sql)
            status    = statement.execute()
            if not status:
                return False

        return True

def register():
    dbpatch.patch.PatchFactory.register(SQLPatch)
