import atheist

import pyodbc


class OdbcConnectionBuilder(object):

    def __init__(self, odbc_driver=None, db_host='localhost',
                 db_user=None, db_pass=None, db_charset='UTF8'):
        self.__check_mandatories([odbc_driver, db_user, db_pass])
        self._driver  = odbc_driver
        self._server  = db_host
        self._user    = db_user
        self._pass    = db_pass
        self._charset = db_charset

    def __check_mandatories(self, mandatories):
        for mandatory in mandatories:
            if mandatory:
                continue
            raise Exception(
                'The odbc_driver[%s], db_user[%s] and ' \
                    'db_pass[%s] arguments are mandatories!!' % \
                    (odbc_driver, db_user, db_pass))


    def create_connection_string(self, db_name=None):
        chain = 'DRIVER={%s};SERVER=%s;USER=%s;PASSWORD=%s;CHARSET=%s;'
        if db_name:
            chain += 'DATABASE=%s;' % db_name

        return chain % (self._driver, self._server, self._user,
                        self._pass, self._charset)


class DbChecker(object):

    def __init__(self, odbc_conn_builder):
        self._odbc_conn_builder = odbc_conn_builder
        self._conn    = None
        self._cursor  = None

    def create_connection_string(self, db_name):
        return self._odbc_conn_builder.create_connection_string(db_name)

    def _connect(self, db_name):
        conn_str = self.create_connection_string(db_name)
        self._conn = pyodbc.connect(conn_str)

    def connect(self, db_name):
        self._connect(db_name)

    def _disconnect(self):
        if self._cursor:
            self._cursor.close()

        if self._conn:
            self._conn.close()

    def disconnect(self):
        self._disconnect()

    def __del__(self):
        self._disconnect()

    def all_databases_query(self):
        return 'SHOW DATABASES'

    def all_database_tables_query(self):
        return 'SHOW TABLES'

    def get_internal_database(self):
        return 'mysql'

    def exist_database(self, db_name):
        all_dbs = self.get_all_databases()
        return db_name in all_dbs

    def exist_table(self, db_name, table):
        self._connect(db_name)
        tables = self.get_all_tables(db_name)
        return table in tables

    def _get_all_entries(self):
        rows = self.get_rows()
        dbs  = []
        for row in rows:
            dbs.append(row[0])

        return dbs

    def get_all_databases(self):
        self._connect(self.get_internal_database())
        self.execute(self.all_databases_query())
        return self._get_all_entries()

    def get_all_tables(self, database):
        self._connect(database)
        self.execute(self.all_database_tables_query())
        return self._get_all_entries()

    def execute(self, query):
        self._cursor = self._conn.execute(query)
        self._conn.commit()

    def get_row(self):
        if not self._cursor:
            return None

        return self._cursor.fetchone()

    def get_rows(self):
        if not self._cursor:
            return []

        return self._cursor.fetchall()

    def get_rows_count(self):
        if not self._cursor:
            return 0

        return self._cursor.rowcount


class MySQLDbChecker(DbChecker):

    def all_databases_query(self):
        return 'SHOW DATABASES'

    def all_database_tables_query(self):
        return 'SHOW TABLES'

    def get_internal_database(self):
        return None

    
class PostgresqlDbChecker(DbChecker):

    def create_connection_string(self, db_name=None):
        base = DbChecker.create_connection_string(self, db_name)
        replacements = {';USER='     : ';UID=',
                        ';PASSWORD=' : ';PWD='}
        for key, value in replacements.items():
            base = base.replace(key, value) 

        return base

    def all_databases_query(self):
        return "SELECT datname FROM pg_database " \
            "WHERE datistemplate = false"

    def all_database_tables_query(self):
        return "SELECT table_name FROM information_schema.tables " \
            "WHERE table_schema = 'public'"

    def get_internal_database(self):
        return 'postgres'

    
class DbCheckerFactory(object):

    @staticmethod
    def create(engine, odbc_conn_builder):
        engine = engine.lower()
        if engine == 'mysql':
            return MySQLDbChecker(odbc_conn_builder)
        elif engine == 'postgresql':
            return PostgresqlDbChecker(odbc_conn_builder)
        else:
            raise Exception("Unsupported engine! [%s]" % engine)
        


class ExistDbTable(atheist.Condition, atheist.Plugin):

    def __init__(self, db_engine=None, db_name=None, db_table=None, 
                 odbc_driver=None, db_host='localhost',
                 db_user=None, db_pass=None, db_charset='UTF8'):
        atheist.Condition.__init__(self)

        self.__check_mandatories({'db_engine' : db_engine,
                                  'db_name'   : db_name,
                                  'db_table'  : db_table,
                                  'odbc_driver' : odbc_driver,
                                  'db_user'   : db_user,
                                  'db_pass'   : db_pass})
        self._db_engine = db_engine
        self._db_name   = db_name
        self._db_table  = db_table
        self._odbc_cb   = OdbcConnectionBuilder(odbc_driver=odbc_driver,
                                                db_user=db_user,
                                                db_pass=db_pass,
                                                db_host=db_host,
                                                db_charset=db_charset)

        self._checker = DbCheckerFactory.create(self._db_engine,
                                                self._odbc_cb)

    def __check_mandatories(self, mandatories):
        for name, value in mandatories.items():
            if value:
                continue
            raise Exception('The argument "%s" is mandatory!!' % name)

    def run(self):
        self._checker.connect(self._db_name)
        status = self._checker.exist_table(self._db_name, self._db_table) 
        return status

    def _base_info(self):
        return '| Check if the table [%s] exists in the database [%s]' % \
            (self._db_table, self._db_name)

    def _full_info(self):
        info = self._base_info()
        info += '. Using "[%s]" db engine' % self._db_engine
        return info

    def basic_info(self):
        return self._full_info()

    def more_info(self):
        return self._full_info()


class ExistDb(atheist.Condition, atheist.Plugin):

    def __init__(self, db_engine=None, db_name=None, 
                 odbc_driver=None, db_host='localhost',
                 db_user=None, db_pass=None, db_charset='UTF8'):
        atheist.Condition.__init__(self)

        self.__check_mandatories({'db_engine' : db_engine,
                                  'db_name'   : db_name,
                                  'odbc_driver' : odbc_driver,
                                  'db_user'   : db_user,
                                  'db_pass'   : db_pass})
        self._db_engine = db_engine
        self._db_name   = db_name
        self._odbc_cb   = OdbcConnectionBuilder(odbc_driver=odbc_driver,
                                                db_user=db_user,
                                                db_pass=db_pass,
                                                db_host=db_host,
                                                db_charset=db_charset)

        self._checker = DbCheckerFactory.create(self._db_engine,
                                                self._odbc_cb)

    def __check_mandatories(self, mandatories):
        for name, value in mandatories.items():
            if value:
                continue
            raise Exception('The argument "%s" is mandatory!!' % name)

    def run(self):
        return self._checker.exist_database(self._db_name)

    def _base_info(self):
        return '| Check if the database [%s] exists' % self._db_name

    def _full_info(self):
        info = self._base_info()
        info += '. Using "[%s]" db engine' % self._db_engine
        return info

    def basic_info(self):
        return self._full_info()

    def more_info(self):
        return self._full_info()


class NumberOfRows(atheist.Condition, atheist.Plugin):

    def __init__(self, db_engine=None, db_name=None, db_table=None,
                 sql_select=None, rows_number=None, odbc_driver=None,
                 db_host='localhost', db_user=None, db_pass=None,
                 db_charset='UTF8'):
        atheist.Condition.__init__(self)

        mandatories = {'db_engine' : db_engine,
                       'db_name'   : db_name,
                       'db_table'  : db_table,
                       'rows_number' : rows_number,
                       'sql_select'  : sql_select,
                       'odbc_driver' : odbc_driver,
                       'db_user'   : db_user,
                       'db_pass'   : db_pass}

        self.__check_mandatories(mandatories)
        self._sql = sql_select
        self._expected_rows = rows_number
        self._db_engine = db_engine
        self._db_name   = db_name
        self._db_table  = db_table
        self._odbc_cb   = OdbcConnectionBuilder(odbc_driver=odbc_driver,
                                                db_user=db_user,
                                                db_pass=db_pass,
                                                db_host=db_host,
                                                db_charset=db_charset)

        self._checker = DbCheckerFactory.create(self._db_engine,
                                                self._odbc_cb)

    def __check_mandatories(self, mandatories):
        for name, value in mandatories.items():
            if value is not None:
                continue
            raise Exception('The argument "%s" is mandatory!!' % name)

    def run(self):
        try:
            self._checker.connect(self._db_name)
            self._checker.execute(self._sql)
        except pyodbc.ProgrammingError, pyErr:
            print '[FAIL] Executing %s test. Error: [%s] ' % \
                (NumberOfRows.__name__, str(pyErr))
            return False

        return self._checker.get_rows_count() == self._expected_rows

    def _base_info(self):
        return '| Check if the table [%s] of [%s] db contains %d rows' % \
            (self._db_table, self._db_name, self._expected_rows)

    def _full_info(self):
        info = self._base_info()
        info += '. Using "[%s]" db engine' % self._db_engine
        return info

    def basic_info(self):
        return self._full_info()

    def more_info(self):
        return self._full_info()


class GreaterNumberOfRows(NumberOfRows):

    def __init__(self, db_engine=None, db_name=None, db_table=None,
                 sql_select=None, rows_number=None,
                 odbc_driver=None, db_host='localhost',
                 db_user=None, db_pass=None, db_charset='UTF8'):
        NumberOfRows.__init__(self, db_engine, db_name, db_table,
                              sql_select, rows_number, odbc_driver,
                              db_host, db_user, db_pass, db_charset)

    def _base_info(self):
        return '| Check if the table [%s] of [%s] db contains ' \
            'more than %d rows' % \
            (self._db_table, self._db_name, self._expected_rows)

    def run(self):
        try:
            self._checker.connect(self._db_name)
            self._checker.execute(self._sql)
        except pyodbc.ProgrammingError, pyErr:
            print '[FAIL] Executing %s test. Error: [%s] ' % \
                (NumberOfRows.__name__, str(pyErr))
            return False

        return self._checker.get_rows_count() > self._expected_rows


class FewerNumberOfRows(NumberOfRows):

    def __init__(self, db_engine=None, db_name=None, db_table=None,
                 sql_select=None, rows_number=None, odbc_driver=None,
                 db_host='localhost', db_user=None, db_pass=None,
                 db_charset='UTF8'):
        NumberOfRows.__init__(self, db_engine, db_name, db_table,
                              sql_select, rows_number, odbc_driver,
                              db_host, db_user, db_pass, db_charset)

    def _base_info(self):
        return '| Check if the table [%s] of [%s] db contains ' \
            'less than %d rows' % \
            (self._db_table, self._db_name, self._expected_rows)

    def run(self):
        try:
            self._checker.connect(self._db_name)
            self._checker.execute(self._sql)
        except pyodbc.ProgrammingError, pyErr:
            print '[FAIL] Executing %s test. Error: [%s] ' % \
                (NumberOfRows.__name__, str(pyErr))
            return False

        return self._checker.get_rows_count() < self._expected_rows
